/*
 * Copyright (c) 2008-2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 ** locoDemoMotionTaskCodels.c
 **
 ** Codels called by execution task locoDemoMotionTask
 **
 ** Author: 
 ** Date: 
 **
 **/

#include <math.h>

#include <portLib.h>
#include <h2mathLib.h>

#include <genPos/genPosStruct.h>
#include <contlaw/contLaw.h>

#include "server/locoDemoHeader.h"

/* Inclusion specifiques */
#include "locoDemoPom.h"
#include "locoDemoConst.h"

/* Pour la simulation des deplacements:
   Memorisation des vitesses simulees du robot */
static struct {
  double v;
  double w;
}speedSimul;

static double prevV = 0;
static double prevW = 0;

/*-------------------- PROTOTYPES DES FONCTIONS LOCALES --------------------*/

static STATUS locoDemoComputConfig   (double  v, double w, int *bilan);
static STATUS locoDemoVelocityGet    (double *v, double *w, int *bilan);
static STATUS locoDemoVelocitySet    (double  v, double  w, int *bilan);


/*------------------------------------------------------------------------
 *
 * initOdoAndAsserv  --  Initialization codel (fIDS, ...)
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
initOdoAndAsserv(int *report)
{
	/* Pointeurs sur les structures a initialiser */
	GENPOS_CART_STATE    *configId;
	GENPOS_CART_REF       *refId;
	GENPOS_CART_CMD_PARAM *cmdId;

	locoDemoDataStrId->mode = LOCODEMO_IDLE;

	/* Initialisation de la configuration */
	configId = &locoDemoDataStrId->robot;
	configId->xRob = 0.;
	configId->yRob = 0.;
	configId->xRef = LOCODEMO_CNTR_POINT_DIST;
	configId->yRef = 0;
	configId->theta = 0.;
	
	/* Initialisation des vitesses */
	configId->v = 0.;
	configId->w = 0.;
	
	/* Initialisation des consignes */
	refId = &locoDemoDataStrId->ref;
	refId->dataType = GENPOS_POS_AND_SPEED_DATA;
	refId->backFlag = GENPOS_FORWARD_MOTION;
	refId->x = LOCODEMO_CNTR_POINT_DIST;
	refId->y = 0.;
	refId->v = 0.;
	refId->w = 0.;
	
	refId->vmax = LOCODEMO_VMAX;
	refId->wmax = LOCODEMO_WMAX;
	refId->linAccelMax = LOCODEMO_ACCEL_LIN_MAX;
	refId->angAccelMax = LOCODEMO_ACCEL_ANG_MAX;

	/* Initialisation des parametres de commandes */
	cmdId = &locoDemoDataStrId->cmd;
	cmdId->distPoint = LOCODEMO_CNTR_POINT_DIST;
	cmdId->KpLongit = LOCODEMO_KP_LONGIT;
	cmdId->KiLongit = LOCODEMO_KI_LONGIT;
	cmdId->KpTransv = LOCODEMO_KP_TRANSV;
	cmdId->KiTransv = LOCODEMO_KI_TRANSV;
	
	cmdId->KpAng = LOCODEMO_KP_ANG;
	cmdId->KiAng = LOCODEMO_KI_ANG;

	//ecart max accepté entre la position et la consigne
	cmdId->longitMaxGap = LOCODEMO_LONGIT_MAX_GAP; 	// if dl > longitMaxGap then tracking return "important drift"(=1)
	cmdId->transvMaxGap = LOCODEMO_TRANSV_MAX_GAP;
	cmdId->angMaxGap    = LOCODEMO_ANG_MAX_GAP;


	//erreurs max raportées (pour l'asservissement proportionnel)
	cmdId->longitErrorMax = LOCODEMO_LONGIT_ERROR_MAX;		//if dl > longitErrorMax then SdL = 0
	cmdId->transvErrorMax = LOCODEMO_TRANSV_ERROR_MAX;		//if dl > longitErrorMax then SdL = 0
	cmdId->angErrorMax =    LOCODEMO_ANG_ERROR_MAX;			//if dl > longitErrorMax then SdL = 0

	//Somme des erreurs max raportées (pour l'asservissement integral)
	cmdId->longitMaxErrSum = LOCODEMO_LONGIT_MAX_ERR_SUM;		//uper bound for SdL
	cmdId->transvMaxErrSum = LOCODEMO_TRANSV_MAX_ERR_SUM;		//uper bound for SdT
	cmdId->angMaxErrSum =    LOCODEMO_ANG_MAX_ERR_SUM;		//uper bound for SdA

	
	/* Initialize Pom */
	initPomPosters();
	
	return(OK);
}

/*------------------------------------------------------------------------
 *
 * odoAndAsserv  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 *              S_locoDemo_EMERGENCY_STOP
 * 
 * Returns:    OK or ERROR
 */

STATUS
odo(int *report)
{
	double v, w;   

	/*
	 * Obtention des vitesses du robot
	 */
	if (locoDemoVelocityGet (&v, &w, report)  != OK)
		return(ERROR);
	
	/*
	 * Calcul de la position
	 */
	if (locoDemoComputConfig (v, w, report) != OK)
		return(ERROR);
	return OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Actual tracking code
 **/
static void
track(int *report, double *vRef, double *wRef)
{
	GENPOS_CART_REF *refId = &locoDemoDataStrId->ref;
	GENPOS_CART_STATE *robot = &locoDemoDataStrId->robot;
	static BOOL prevDrifted = FALSE;
	BOOL drifted;

	/* 
	 * Calcul des vitesses d'asservissement
	 */
	switch (refId->dataType) {
	case GENPOS_POS_DATA:
		drifted = contLawCartPositionControl(&locoDemoDataStrId->cmd,
		    refId, robot, EXEC_TASK_PERIOD(LOCODEMO_MOTIONTASK_NUM),
		    prevV, prevW, vRef, wRef);
		break;
	case GENPOS_POS_AND_SPEED_DATA:
		drifted = contLawCartConfigControl(&locoDemoDataStrId->cmd,
			&locoDemoDataStrId->ref,
			&locoDemoDataStrId->robot,
			EXEC_TASK_PERIOD(LOCODEMO_MOTIONTASK_NUM),
		    prevV, prevW, vRef, wRef);
		break;
	case GENPOS_SPEED_DATA:
		*vRef = (refId->backFlag == GENPOS_FORWARD_MOTION ? refId->v :
		    -refId->v);
		*wRef = refId->w;
		break;
	default:
		*report = S_locoDemo_BAD_REF;
		break;
	}
	if (drifted && !prevDrifted) {
		printf("track: important drift!\n");
		printf("ref: %.2lf %.2lf %.2lf\n",
		       refId->x, refId->y, refId->theta);
		printf("pos: %.2lf %.2lf %.2lf\n",
		       robot->xRef, robot->yRef, robot->theta);
	}
	if (!drifted && prevDrifted)
		printf("track: drift cancelled\n");
	prevDrifted = drifted;
	prevV = *vRef;
	prevW = *wRef;
}

/*------------------------------------------------------------------------
 *
 * Asserv  --  2nd codel of permanent activity (called AFTER other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 *              S_locoDemo_EMERGENCY_STOP
 * 
 * Returns:    OK or ERROR
 */

STATUS
asserv(int *report)
{
	GENPOS_CART_REF *refId = &locoDemoDataStrId->ref;
	GENPOS_CART_STATE *robot = &locoDemoDataStrId->robot;
	double vRef, wRef;
	
	*report = OK;
	switch (locoDemoDataStrId->mode) {
	case LOCODEMO_MOTORS_OFF:
	case LOCODEMO_EMERGENCY:
		/* No motion possible */
		return OK;
		
	case LOCODEMO_IDLE:
		vRef = 0.0;
		wRef = 0.0;
		refId->x = robot->xRef;
		refId->y = robot->yRef;
		refId->theta = robot->theta;
		break;
		
	case LOCODEMO_MANUAL:
		/* Not implemented */
		return OK;

	case LOCODEMO_TRACK:
		track(report, &vRef, &wRef);
		break;

	default:
		*report = S_locoDemo_BAD_REF;
		return ERROR;
	}
	if (*report != OK) {
		locoDemoDataStrId->mode = LOCODEMO_IDLE;
		locoDemoDataStrId->trackStr.trackMode = GENPOS_NO_TRACKING;
		vRef = 0;
		wRef = 0;
	}
	if (locoDemoVelocitySet (vRef, wRef, report) != OK)
		return(ERROR);
	
	updatePomPosters();
	
	return(OK);
}

/*------------------------------------------------------------------------
 * SurvLine
 *
 * Description: 
 *
 * Reports:      OK
 */

/* survLine  -  codel EXEC of SurvLine
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
survLine(SURV_LINE_STR *survLineStr, int *report)
{
	double xLine, yLine;       /* Vecteur directeur de la droite */
	double nLine;              /* Norme du vecteur */
	double xPtSeg, yPtSeg;     /* Vecteur 1er point- pt surveille */
	double distPtLine;         /* Distance point-ligne */
	
	/* Vecteur directeur de la droite */
	xLine =  survLineStr->pt2[0] - survLineStr->pt1[0];
	yLine =  survLineStr->pt2[1] - survLineStr->pt1[1];
	nLine = sqrt(xLine * xLine + yLine * yLine);
	
	/* Vecteur 1er point - robot */
	xPtSeg = locoDemoDataStrId->robot.xRob - survLineStr->pt1[0];
	yPtSeg = locoDemoDataStrId->robot.yRob - survLineStr->pt1[1];
	
	/* Distance orientee point surveille-ligne:
	   pas passee: negative,  passe: positive */
	distPtLine = (xLine * yPtSeg - yLine * xPtSeg)/nLine;
	
	/* Point surveille au dessus de la ligne */
	if (distPtLine > 0)
		return(END);
	
	/* En dessous de la ligne */
	else
		return(EXEC);
}


/*------ FONCTIONS AUXILIAIRES de odoAndAsserv ---------------------------*/



/*****************************************************************************
 *
 *  locoDemoComputConfig  -  Calcul la configuration du robot.
 *
 *  Description: Calcul incremental de la configuration du robot a partir 
 *               des vitesses lineaire et angulaire instantannees.
 *
 *  Retourne: OK ou ERROR
 */

static STATUS locoDemoComputConfig (double v,
				    double w,
				    int *bilan)
{
	double dS, dT;          /* Variations lineaires et angulaires */
	double thetaMoy;        /* Angle moyen entre 2 config successives */
	double c, s;            /* Cos et sin de l'angle moyen */
	double locoDemoPeriod = EXEC_TASK_PERIOD(LOCODEMO_MOTIONTASK_NUM);/* Periode */
	GENPOS_CART_STATE *posId = &(locoDemoDataStrId->robot);
	GENPOS_CART_REF *refId = &(locoDemoDataStrId->ref);
	int direction;
	
	/* Variations lineaires et angulaires */
	dS = v * locoDemoPeriod;
	dT = w * locoDemoPeriod;
	
	/* Vitesses du robot */
	posId->v = v;
	posId->w = w;
	
	/* Configuration du robot */
	thetaMoy = posId->theta + dT / 2;
	c = cos(thetaMoy);
	s = sin(thetaMoy);
	posId->theta = angleLimit(posId->theta + dT);
	posId->xRob += dS * c;
	posId->yRob += dS * s;
	
	if (refId->backFlag == GENPOS_FORWARD_MOTION)
		direction = 1;
	else
		direction = -1;
	posId->xRef = posId->xRob + direction*
	    LOCODEMO_CNTR_POINT_DIST*cos(posId->theta);
	posId->yRef = posId->yRob + direction*
	    LOCODEMO_CNTR_POINT_DIST*sin(posId->theta);
	return(OK);
}

/****************************************************************************
 *
 *  locoDemoVelocityGet  -  Obtention des vitesses du robot.
 *
 *  Description: Simule la lecture odometrique des vitesses 
 *               lineaire et angulaire du robot.
 *
 *  Retourne: OK ou ERROR
 */

static STATUS locoDemoVelocityGet (double *v,    /* Vitesse lin. du robot */
				   double *w,    /* Vitesse ang. du robot */
				   int *bilan) 
{
	*v = speedSimul.v;
	*w = speedSimul.w;
	return(OK);
}



/****************************************************************************
 *
 *  locoDemoVelocitySet  -  Commande en vitesses du robot.
 *
 *  Description: Simule la commande en vitesses du robot.
 *               Hypothese: Le robot suit parfaitement la consigne en 
 *               vitesses (les limites en vitesses et acceleration on deja 
 *               ete limitees lors du calcul de la consigne).
 *
 *  Retourne: OK ou ERROR
 */

static STATUS locoDemoVelocitySet (double v,  /* Vitesse lin. commande' */
				   double w,  /* Vitesse ang. commande' */
				   int *bilan)
{
	speedSimul.v = v;
	speedSimul.w = w;
	return(OK);
}

